@@ -1,29 +1,35 @@
 import logging
+import re
 
 from inspect import Parameter
-from typing import List, Dict, Optional, Any
+from typing import List, Dict, Optional, Any, Iterator, Callable, Union
 
-from jedi import (Script, create_environment, get_default_environment,
+from jedi import (Script, create_environment,  # type: ignore
+                  get_default_environment,
                   settings as jedi_settings, get_default_project)
-from jedi.api.classes import Name
+from jedi.api.classes import Name, Completion  # type: ignore
 
-from pycodestyle import (BaseReport as CodestyleBaseReport,
+from pycodestyle import (BaseReport as CodestyleBaseReport,  # type: ignore
                          Checker as CodestyleChecker,
                          StyleGuide as CodestyleStyleGuide)
 
-from pyflakes.api import check as pyflakes_check
+from pyflakes.api import check as pyflakes_check  # type: ignore
 
 from pygls.features import (COMPLETION, TEXT_DOCUMENT_DID_CHANGE,
                             TEXT_DOCUMENT_DID_CLOSE, TEXT_DOCUMENT_DID_OPEN,
                             HOVER, SIGNATURE_HELP, DEFINITION,
                             REFERENCES, WORKSPACE_DID_CHANGE_CONFIGURATION,
-                            TEXT_DOCUMENT_WILL_SAVE, TEXT_DOCUMENT_DID_SAVE)
+                            TEXT_DOCUMENT_WILL_SAVE, TEXT_DOCUMENT_DID_SAVE,
+                            DOCUMENT_SYMBOL)
 from pygls import types
 from pygls.server import LanguageServer
 from pygls.protocol import LanguageServerProtocol
 from pygls.uris import from_fs_path, to_fs_path
 
 
+RE_WORD = re.compile(r'\w*')
+
+
 _COMPLETION_TYPES = {
     'module': types.CompletionItemKind.Module,
     'class': types.CompletionItemKind.Class,
@@ -38,14 +44,22 @@ _COMPLETION_TYPES = {
 jedi_settings.case_insensitive_completion = False
 
 
+completionFunction: Callable[[List[Completion], types.Range],
+                             Iterator[types.CompletionItem]]
+documentSymbolFunction: Union[
+    Callable[[str, List[str], List[Name]], List[types.DocumentSymbol]],
+    Callable[[str, List[str], List[Name]], List[types.SymbolInformation]]]
+
+
 class AnakinLanguageServerProtocol(LanguageServerProtocol):
 
     def bf_initialize(
             self, params: types.InitializeParams) -> types.InitializeResult:
         result = super().bf_initialize(params)
-
         global jediEnvironment
         global jediProject
+        global completionFunction
+        global documentSymbolFunction
         venv = getattr(params.initializationOptions, 'venv', None)
         if venv:
             jediEnvironment = create_environment(venv, False)
@@ -58,6 +72,27 @@ class AnakinLanguageServerProtocol(LanguageServerProtocol):
             logging.info(f'  {p}')
         logging.info(f'Jedi project path: {jediProject._path}')
 
+        def get_attr(o, *attrs):
+            try:
+                for attr in attrs:
+                    o = getattr(o, attr)
+                return o
+            except AttributeError:
+                return None
+
+        caps = getattr(params.capabilities, 'textDocument', None)
+
+        if get_attr(caps, 'completion', 'completionItem', 'snippetSupport'):
+            completionFunction = _completions_snippets
+        else:
+            completionFunction = _completions
+
+        if get_attr(caps,
+                    'documentSymbol', 'hierarchicalDocumentSymbolSupport'):
+            documentSymbolFunction = _document_symbol_hierarchy
+        else:
+            documentSymbolFunction = _document_symbol_plain
+
         result.capabilities.textDocumentSync = types.TextDocumentSyncOptions(
             open_close=True,
             change=types.TextDocumentSyncKind.INCREMENTAL,
@@ -67,11 +102,14 @@ class AnakinLanguageServerProtocol(LanguageServerProtocol):
 
 
 server = LanguageServer(protocol_cls=AnakinLanguageServerProtocol)
+
 scripts: Dict[str, Script] = {}
 pycodestyleOptions: Dict[str, Any] = {}
 mypyConfigs: Dict[str, str] = {}
+
 jediEnvironment = None
 jediProject = None
+
 config = {
     'pyflakes_errors': [
         'UndefinedName'
@@ -317,6 +355,74 @@ def did_change(ls: LanguageServer, params: types.DidChangeTextDocumentParams):
     get_script(ls, params.textDocument.uri, True)
 
 
+def _completion_sort_key(completion: Completion) -> str:
+    name = completion.name
+    if name.startswith('__'):
+        return f'zz{name}'
+    if name.startswith('_'):
+        return f'za{name}'
+    return f'aa{name}'
+
+
+def _completion_item(completion: Completion, r: types.Range) -> Dict:
+    label = completion.name
+    if not completion.complete.startswith("'") and label.startswith("'"):
+        label = label[1:]
+    return dict(
+        label=label,
+        kind=_COMPLETION_TYPES.get(completion.type,
+                                   types.CompletionItemKind.Text),
+        documentation=completion.docstring(raw=True),
+        sort_text=_completion_sort_key(completion),
+        text_edit=types.TextEdit(r, completion.complete)
+    )
+
+
+def _completions(completions: List[Completion],
+                 r: types.Range) -> Iterator[types.CompletionItem]:
+    return (
+        types.CompletionItem(
+            **_completion_item(completion, r)
+        ) for completion in completions
+    )
+
+
+def _completions_snippets(completions: List[Completion],
+                          r: types.Range) -> Iterator[types.CompletionItem]:
+    for completion in completions:
+        item = _completion_item(completion, r)
+        yield types.CompletionItem(
+            **item
+        )
+        for signature in completion.get_signatures():
+            names = []
+            snippets = []
+            for i, param in enumerate(signature.params):
+                if param.kind == Parameter.VAR_KEYWORD:
+                    break
+                if '=' in param.description:
+                    break
+                if param.name == '/':
+                    continue
+                names.append(param.name)
+                if param.kind == Parameter.KEYWORD_ONLY:
+                    snippet_prefix = f'{param.name}='
+                else:
+                    snippet_prefix = ''
+                snippets.append(
+                    f'{snippet_prefix}${{{i + 1}:{param.name}}}'
+                )
+            names_str = ', '.join(names)
+            snippets_str = ', '.join(snippets)
+            yield types.CompletionItem(**dict(
+                item,
+                label=f'{completion.name}({names_str})',
+                insert_text=f'{completion.name}({snippets_str})$0',
+                insert_text_format=types.InsertTextFormat.Snippet,
+                text_edit=None
+            ))
+
+
 @server.feature(COMPLETION, trigger_characters=['.'])
 def completions(ls: LanguageServer, params: types.CompletionParams):
     script = get_script(ls, params.textDocument.uri)
@@ -324,59 +430,20 @@ def completions(ls: LanguageServer, params: types.CompletionParams):
         params.position.line + 1,
         params.position.character
     )
-    position = types.Position(params.position.line,
-                              params.position.character)
-    r = types.Range(position, position)
-
-    def sort_key(completion):
-        name = completion.name
-        if name.startswith('__'):
-            return f'zz{name}'
-        if name.startswith('_'):
-            return f'za{name}'
-        return f'aa{name}'
-
-    def _completions():
-        for completion in completions:
-            item = dict(
-                label=completion.name,
-                kind=_COMPLETION_TYPES.get(completion.type,
-                                           types.CompletionItemKind.Text),
-                documentation=completion.docstring(raw=True),
-                sort_text=sort_key(completion)
-            )
-            yield types.CompletionItem(
-                text_edit=types.TextEdit(r, completion.complete),
-                **item
-            )
-            for signature in completion.get_signatures():
-                names = []
-                snippets = []
-                for i, param in enumerate(signature.params):
-                    if param.kind == Parameter.VAR_KEYWORD:
-                        break
-                    if '=' in param.description:
-                        break
-                    if param.name == '/':
-                        continue
-                    names.append(param.name)
-                    if param.kind == Parameter.KEYWORD_ONLY:
-                        snippet_prefix = f'{param.name}='
-                    else:
-                        snippet_prefix = ''
-                    snippets.append(
-                        f'{snippet_prefix}${{{i + 1}:{param.name}}}'
-                    )
-                names_str = ', '.join(names)
-                snippets_str = ', '.join(snippets)
-                yield types.CompletionItem(**dict(
-                    item,
-                    label=f'{completion.name}({names_str})',
-                    insert_text=f'{completion.name}({snippets_str})$0',
-                    insert_text_format=types.InsertTextFormat.Snippet
-                ))
-
-    return types.CompletionList(False, list(_completions()))
+    code_line = script._code_lines[params.position.line]
+    word_match = RE_WORD.match(code_line[params.position.character:])
+    if word_match:
+        word_rest = word_match.end()
+    else:
+        word_rest = 0
+    r = types.Range(
+        types.Position(params.position.line,
+                       params.position.character),
+        types.Position(params.position.line,
+                       params.position.character + word_rest)
+    )
+    return types.CompletionList(False,
+                                list(completionFunction(completions, r)))
 
 
 @server.feature(HOVER)
@@ -484,3 +551,97 @@ def will_save(ls: LanguageServer, params: types.WillSaveTextDocumentParams):
 @server.feature(TEXT_DOCUMENT_DID_SAVE)
 def did_save(ls: LanguageServer, params: types.DidSaveTextDocumentParams):
     _validate(ls, params.textDocument.uri)
+
+
+_DOCUMENT_SYMBOL_KINDS = {
+    'module': types.SymbolKind.Module,
+    'class': types.SymbolKind.Class,
+    'function': types.SymbolKind.Function,
+    'statement': types.SymbolKind.Variable,
+    'instance': types.SymbolKind.Variable,
+    '_pseudotreenameclass': types.SymbolKind.Class
+}
+
+
+def _get_document_symbols(
+        code_lines: List[str],
+        names: List[Name],
+        current: Optional[Name] = None
+) -> List[types.DocumentSymbol]:
+    # Looks like names are sorted by order of appearance, so
+    # children are after their parents
+    result = []
+    while names:
+        if current and names[0].parent() != current:
+            break
+        name = names.pop(0)
+        if name.type == 'param':
+            continue
+        children = _get_document_symbols(
+            code_lines,
+            names,
+            name
+        )
+        line = name.line - 1
+        r = types.Range(
+            types.Position(line, name.column),
+            types.Position(line, len(code_lines[line]) - 1)
+        )
+        result.append(types.DocumentSymbol(
+            name.name,
+            _DOCUMENT_SYMBOL_KINDS.get(name.type, types.SymbolKind.Null),
+            r,
+            r,
+            children=children or None
+        ))
+    return result
+
+
+def _document_symbol_hierarchy(
+        uri: str, code_lines: List[str], names: List[Name]
+) -> List[types.DocumentSymbol]:
+    return _get_document_symbols(code_lines, names)
+
+
+def _document_symbol_plain(
+        uri: str, code_lines: List[str], names: List[Name]
+) -> List[types.SymbolInformation]:
+    def _symbols():
+        for name in names:
+            if name.type == 'param':
+                continue
+            parent = name.parent()
+            parent_name = parent and parent.full_name
+            if parent_name:
+                module_name = name.module_name
+                if parent_name == module_name:
+                    parent_name = None
+                elif parent_name.startswith(f'{module_name}.'):
+                    parent_name = parent_name[len(module_name) + 1:]
+            yield types.SymbolInformation(
+                name.name,
+                _DOCUMENT_SYMBOL_KINDS.get(name.type, types.SymbolKind.Null),
+                types.Location(uri, types.Range(
+                    types.Position(name.line - 1, name.column),
+                    types.Position(name.line - 1,
+                                   len(code_lines[name.line - 1]) - 1)
+                )),
+                parent_name
+            )
+    return list(_symbols())
+
+
+@server.feature(DOCUMENT_SYMBOL)
+def document_symbol(
+        ls: LanguageServer, params: types.DocumentSymbolParams
+) -> Union[List[types.DocumentSymbol], List[types.SymbolInformation], None]:
+    script = get_script(ls, params.textDocument.uri)
+    names = script.get_names(all_scopes=True)
+    if not names:
+        return None
+    result = documentSymbolFunction(
+        params.textDocument.uri,
+        script._code_lines,
+        script.get_names(all_scopes=True)
+    )
+    return result
